# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings
import dddp
import dddp.models
from dddp.migrations import TruncateOperation


class Migration(migrations.Migration):

    dependencies = [
        ('sessions', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dddp', '0001_initial'),
    ]

    operations = [
        TruncateOperation(forwards=['subscription']),
        migrations.CreateModel(
            name='Connection',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('connection_id', dddp.models.AleaIdField(default=dddp.meteor_random_id, max_length=17)),
                ('remote_addr', models.CharField(max_length=255)),
                ('version', models.CharField(max_length=255)),
                ('session', models.ForeignKey(to='sessions.Session')),
            ],
            options={
            },
            bases=(models.Model, object),
        ),
        migrations.CreateModel(
            name='SubscriptionCollection',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('collection_class', models.CharField(max_length=255)),
                ('subscription', models.ForeignKey(related_name='collections', to='dddp.Subscription')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='connection',
            unique_together=set([('connection_id', 'session')]),
        ),
        migrations.AddField(
            model_name='subscription',
            name='connection',
            field=models.ForeignKey(to='dddp.Connection'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='subscription',
            name='publication_class',
            field=models.CharField(max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='subscription',
            name='sub_id',
            field=models.CharField(max_length=17),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='subscription',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='subscription',
            unique_together=set([('connection', 'sub_id')]),
        ),
        migrations.RemoveField(
            model_name='subscription',
            name='session',
        ),
        TruncateOperation(backwards=['subscription'])
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import dddp.models


class Migration(migrations.Migration):

    dependencies = [
        ('sessions', '0001_initial'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ObjectMapping',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('meteor_id', dddp.models.AleaIdField(max_length=17)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('publication', models.CharField(max_length=255)),
                ('params_ejson', models.TextField(default=b'{}')),
                ('session', models.ForeignKey(to='sessions.Session')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='objectmapping',
            unique_together=set([('content_type', 'meteor_id')]),
        ),
        migrations.AlterIndexTogether(
            name='objectmapping',
            index_together=set([('content_type', 'meteor_id'), ('content_type', 'object_id')]),
        ),
    ]



"""Django DDP Server app config."""
from __future__ import print_function, absolute_import, unicode_literals

import io
import os.path

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from ejson import dumps, loads
import pybars


STAR_JSON_SETTING_NAME = 'METEOR_STAR_JSON'


def read(path, default=None, encoding='utf8'):
    """Read encoded contents from specified path or return default."""
    if not path:
        return default
    try:
        with io.open(path, mode='r', encoding=encoding) as contents:
            return contents.read()
    except IOError:
        if default is not None:
            return default
        raise


def read_json(path):
    """Read JSON encoded contents from specified path."""
    with open(path, mode='r') as json_file:
        return loads(json_file.read())


class ServerConfig(AppConfig):

    """Django config for dddp.server app."""

    name = 'dddp.server'
    verbose_name = 'Django DDP Meteor Web Server'

    manifest = None
    program_json = None
    program_json_path = None
    runtime_config = None

    star_json = None  # top level layout
    server_json = None  # web server layout
    web_browser_json = None  # web.browser (client) layout

    url_map = None
    internal_map = None
    server_load_map = None
    template_path = None  # web server HTML template path
    client_map = None  # web.browser (client) URL to path map
    html = '<!DOCTYPE html>\n<html><head><title>DDP App</title></head></html>'

    def ready(self):
        """Configure Django DDP server app."""
        self.url_map = {}

        try:
            json_path = getattr(settings, STAR_JSON_SETTING_NAME)
        except AttributeError:
            raise ImproperlyConfigured(
                '%s setting required by dddp.server.view.' % (
                    STAR_JSON_SETTING_NAME,
                ),
            )

        self.star_json = read_json(json_path)
        star_format = self.star_json['format']
        if star_format != 'site-archive-pre1':
            raise ValueError(
                'Unknown Meteor star format: %r' % star_format,
            )
        programs = {
            program['name']: program
            for program in self.star_json['programs']
        }

        server_json_path = os.path.join(
            os.path.dirname(json_path),
            os.path.dirname(programs['server']['path']),
            'program.json',
        )
        self.server_json = read_json(server_json_path)
        server_format = self.server_json['format']
        if server_format != 'javascript-image-pre1':
            raise ValueError(
                'Unknown Meteor server format: %r' % server_format,
            )
        self.server_load_map = {}
        for item in self.server_json['load']:
            item['path_full'] = os.path.join(
                os.path.dirname(server_json_path),
                item['path'],
            )
            self.server_load_map[item['path']] = item
            self.url_map[item['path']] = (
                item['path_full'], 'text/javascript'
            )
            try:
                item['source_map_full'] = os.path.join(
                    os.path.dirname(server_json_path),
                    item['sourceMap'],
                )
                self.url_map[item['sourceMap']] = (
                    item['source_map_full'], 'text/plain'
                )
            except KeyError:
                pass
        self.template_path = os.path.join(
            os.path.dirname(server_json_path),
            self.server_load_map[
                'packages/boilerplate-generator.js'
            ][
                'assets'
            ][
                'boilerplate_web.browser.html'
            ],
        )

        web_browser_json_path = os.path.join(
            os.path.dirname(json_path),
            programs['web.browser']['path'],
        )
        self.web_browser_json = read_json(web_browser_json_path)
        web_browser_format = self.web_browser_json['format']
        if web_browser_format != 'web-program-pre1':
            raise ValueError(
                'Unknown Meteor web.browser format: %r' % (
                    web_browser_format,
                ),
            )
        self.client_map = {}
        self.internal_map = {}
        for item in self.web_browser_json['manifest']:
            item['path_full'] = os.path.join(
                os.path.dirname(web_browser_json_path),
                item['path'],
            )
            if item['where'] == 'client':
                if '?' in item['url']:
                    item['url'] = item['url'].split('?', 1)[0]
                self.client_map[item['url']] = item
                self.url_map[item['url']] = (
                    item['path_full'],
                    {
                        'js': 'text/javascript',
                        'css': 'text/css',
                    }.get(
                        item['type'], 'application/octet-stream',
                    ),
                )
            elif item['where'] == 'internal':
                self.internal_map[item['type']] = item

        config = {
            'css': [
                {'url': item['path']}
                for item in self.web_browser_json['manifest']
                if item['type'] == 'css' and item['where'] == 'client'
            ],
            'js': [
                {'url': item['path']}
                for item in self.web_browser_json['manifest']
                if item['type'] == 'js' and item['where'] == 'client'
            ],
            'meteorRuntimeConfig': '"%s"' % (
                dumps(self.runtime_config)
            ),
            'rootUrlPathPrefix': '/app',
            'bundledJsCssPrefix': '/app/',
            'inlineScriptsAllowed': False,
            'inline': None,
            'head': read(
                self.internal_map.get('head', {}).get('path_full', None),
                default=u'',
            ),
            'body': read(
                self.internal_map.get('body', {}).get('path_full', None),
                default=u'',
            ),
        }
        tmpl_raw = read(self.template_path, encoding='utf8')
        compiler = pybars.Compiler()
        tmpl = compiler.compile(tmpl_raw)
        self.html = '<!DOCTYPE html>\n%s' % tmpl(config)



"""Django DDP models."""
from __future__ import absolute_import

import collections

from django.db import models, transaction
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.utils.encoding import python_2_unicode_compatible
import ejson
from dddp import meteor_random_id


@transaction.atomic
def get_meteor_id(obj_or_model, obj_pk=None):
    """Return an Alea ID for the given object."""
    if obj_or_model is None:
        return None
    # Django model._meta is now public API -> pylint: disable=W0212
    meta = obj_or_model._meta
    model = meta.model
    if model is ObjectMapping:
        # this doesn't make sense - raise TypeError
        raise TypeError("Can't map ObjectMapping instances through self.")
    if obj_or_model is not model and obj_pk is None:
        obj_pk = str(obj_or_model.pk)
    if obj_pk is None:
        return None
    content_type = ContentType.objects.get_for_model(model)
    try:
        return ObjectMapping.objects.values_list(
            'meteor_id', flat=True,
        ).get(
            content_type=content_type,
            object_id=obj_pk,
        )
    except ObjectDoesNotExist:
        return ObjectMapping.objects.create(
            content_type=content_type,
            object_id=obj_pk,
            meteor_id=meteor_random_id('/collection/%s' % meta),
        ).meteor_id


@transaction.atomic
def get_meteor_ids(model, object_ids):
    """Return Alea ID mapping for all given ids of specified model."""
    content_type = ContentType.objects.get_for_model(model)
    result = collections.OrderedDict(
        (str(obj_pk), None)
        for obj_pk
        in object_ids
    )
    for obj_pk, meteor_id in ObjectMapping.objects.filter(
            content_type=content_type,
            object_id__in=list(result)
    ).values_list('object_id', 'meteor_id'):
        result[obj_pk] = meteor_id
    for obj_pk, meteor_id in result.items():
        if meteor_id is None:
            # Django model._meta is now public API -> pylint: disable=W0212
            result[obj_pk] = ObjectMapping.objects.create(
                content_type=content_type,
                object_id=obj_pk,
                meteor_id=meteor_random_id('/collection/%s' % model._meta),
            ).meteor_id
    return result


@transaction.atomic
def get_object_id(model, meteor_id):
    """Return an object ID for the given meteor_id."""
    if model is ObjectMapping:
        # this doesn't make sense - raise TypeError
        raise TypeError("Can't map ObjectMapping instances through self.")
    # Django model._meta is now public API -> pylint: disable=W0212
    content_type = ContentType.objects.get_for_model(model)
    return ObjectMapping.objects.filter(
        content_type=content_type,
        meteor_id=meteor_id,
    ).values_list('object_id', flat=True).get()


@transaction.atomic
def get_object(model, meteor_id, *args, **kwargs):
    """Return an object for the given meteor_id."""
    return model.objects.filter(*args, **kwargs).get(
        pk=get_object_id(model, meteor_id),
    )


class AleaIdField(models.CharField):

    """CharField that generates its own values using Alea PRNG before INSERT."""

    def __init__(self, *args, **kwargs):
        """Assume max_length of 17 to match Meteor implementation."""
        kwargs.update(
            default=meteor_random_id,
            max_length=17,
        )
        super(AleaIdField, self).__init__(*args, **kwargs)


@python_2_unicode_compatible
class ObjectMapping(models.Model):

    """Mapping from regular Django model primary keys to Meteor object IDs."""

    meteor_id = AleaIdField()
    content_type = models.ForeignKey(ContentType, db_index=True)
    object_id = models.CharField(max_length=255)
    # content_object = GenericForeignKey('content_type', 'object_id')

    def __str__(self):
        """Text representation of a mapping."""
        return '%s: %s[%s]' % (
            self.meteor_id, self.content_type, self.object_id,
        )

    class Meta(object):

        """Meta info for ObjectMapping model."""

        unique_together = [
            ['content_type', 'meteor_id'],
        ]
        index_together = [
            ['content_type', 'object_id'],
            ['content_type', 'meteor_id'],
        ]


@python_2_unicode_compatible
class Connection(models.Model, object):

    """Django DDP connection instance."""

    connection_id = AleaIdField()
    server_addr = models.CharField(max_length=255)
    remote_addr = models.CharField(max_length=255)
    version = models.CharField(max_length=255)

    def __str__(self):
        """Text representation of subscription."""
        return u'%s/\u200b%s' % (
            self.connection_id,
            self.remote_addr,
        )


@python_2_unicode_compatible
class Subscription(models.Model, object):

    """Subscription to a publication with params."""

    _publication_cache = {}
    connection = models.ForeignKey(Connection)
    sub_id = models.CharField(max_length=17)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True)
    publication = models.CharField(max_length=255)
    params_ejson = models.TextField(default='{}')

    class Meta(object):

        """Subscription model meta."""

        unique_together = [
            ['connection', 'sub_id'],
        ]

    def __str__(self):
        """Text representation of subscription."""
        return u'%s/\u200b%s/\u200b%s: %s%s' % (
            self.user,
            self.connection_id,
            self.sub_id,
            self.publication,
            self.params_ejson,
        )

    def get_params(self):
        """Get params dict."""
        return ejson.loads(self.params_ejson or '{}')

    def set_params(self, vals):
        """Set params dict."""
        self.params_ejson = ejson.dumps(vals or {})

    params = property(get_params, set_params)


@python_2_unicode_compatible
class SubscriptionCollection(models.Model):

    """Collections for a subscription."""

    subscription = models.ForeignKey(Subscription, related_name='collections')
    model_name = models.CharField(max_length=255)
    collection_name = models.CharField(max_length=255)

    def __str__(self):
        """Human readable representation of colleciton for a subscription."""
        return u'%s \u200b %s (%s)' % (
            self.subscription,
            self.collection_name,
            self.model_name,
        )



from __future__ import unicode_literals
from django.contrib import admin
from django.core.urlresolvers import reverse, NoReverseMatch
from django.utils.html import format_html
from dddp import models


def object_admin_link(obj):
    kwargs = {
        'format_string': u'{app_label}.{model} {object_id}: {object}',
        'app_label': obj.content_type.app_label,
        'model': obj.content_type.model,
        'object_id': obj.object_id,
        'object': obj.content_type.model_class().objects.get(pk=obj.object_id),
    }
    try:
        kwargs.update(
            url=reverse(
                'admin:%s_%s_change' % (
                    obj.content_type.app_label,
                    obj.content_type.model,
                ),
                args=(obj.object_id,),
            ),
            format_string='<a href="{url}">%s</a>' % kwargs['format_string'],
        )
    except NoReverseMatch:
        pass
    return format_html(**kwargs)
object_admin_link.short_description = 'Object'
object_admin_link.allow_tags = True


class ObjectMapping(admin.ModelAdmin):
    search_fields = [
        'meteor_id',
    ]
    list_display = [
        'meteor_id',
        object_admin_link,
        'content_type',
        'object_id',
    ]
    list_filter = [
        'content_type__app_label',
        'content_type',
    ]
    list_select_related = [
        'content_type',
    ]


class SubscriptionCollectionInline(admin.TabularInline):
    model = models.SubscriptionCollection
    fields = [
        'model_name',
        'collection_name',
    ]
    readonly_fields = [
        'model_name',
        'collection_name',
    ]
    max_num = 0


class Subscription(admin.ModelAdmin):
    search_fields = [
        'sub_id',
        'publication',
        'publication_class',
        'params_ejson',
    ]
    list_display = [
        'sub_id',
        'connection',
        'user',
        'publication',
        'params_ejson',
    ]
    list_filter = [
        'user__email',
        'publication',
    ]
    inlines = [
        SubscriptionCollectionInline,
    ]


class SubscriptionCollection(admin.ModelAdmin):
    search_fields = [
        'model_name',
        'collection_name',
    ]
    list_display = [
        '__str__',
        'subscription',
        'model_name',
        'collection_name',
    ]
    list_filter = [
        'model_name',
        'collection_name',
    ]


for name, attr in vars(models).items():
    if hasattr(attr, '_meta'):
        model_admin = locals().get(name, None)
        if model_admin is not False:
            admin.site.register(attr, model_admin)



from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Task(models.Model):
    text = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ['created_at']

    def __str__(self):
        return self.text



"""
Django DDP authentication.

Matches Meteor 1.1 Accounts package: https://www.meteor.com/accounts

See http://docs.meteor.com/#/full/accounts_api for details of each method.
"""
from binascii import Error
import collections

from ejson import loads, dumps

from django.contrib import auth
from django.contrib.sessions.backends.db import SessionStore
from django.contrib.auth.signals import user_login_failed
from django.dispatch import Signal
from django.utils import timezone

from dddp import THREAD_LOCAL as this, ADDED, REMOVED
from dddp.models import get_meteor_id, get_object, Subscription
from dddp.api import API, APIMixin, api_endpoint, Collection, Publication
from dddp.websocket import MeteorError


create_user = Signal(providing_args=['request', 'params'])
password_changed = Signal(providing_args=['request', 'user'])
forgot_password = Signal(providing_args=['request', 'user', 'token', 'expiry'])
password_reset = Signal(providing_args=['request', 'user'])


class Users(Collection):

    """Mimic `users` collection of Meteor's `accounts-password` package."""

    name = 'users'
    api_path_prefix = '/users/'
    model = auth.get_user_model()

    user_rel = [
        'pk',
    ]

    def serialize(self, obj, *args, **kwargs):
        """Serialize user as per Meteor accounts serialization."""
        # use default serialization, then modify to suit our needs.
        data = super(Users, self).serialize(obj, *args, **kwargs)

        # everything that isn't handled explicitly ends up in `profile`
        profile = data.pop('fields')
        profile.setdefault('name', obj.get_full_name())
        fields = data['fields'] = {
            'username': obj.get_username(),
            'emails': [],
            'profile': profile,
        }

        # clear out sensitive data
        for sensitive in [
                'password',
                'user_permissions_ids',
                'is_active',
                'is_staff',
                'is_superuser',
                'groups_ids',
        ]:
            profile.pop(sensitive, None)

        # createdAt (default is django.contrib.auth.models.User.date_joined)
        try:
            fields['createdAt'] = profile.pop('date_joined')
        except KeyError:
            date_joined = getattr(
                obj, 'get_date_joined',
                lambda: getattr(obj, 'date_joined', None)
            )()
            if date_joined:
                fields['createdAt'] = date_joined

        # email (default is django.contrib.auth.models.User.email)
        try:
            email = profile.pop('email')
        except KeyError:
            email = getattr(
                obj, 'get_email',
                lambda: getattr(obj, 'email', None)
            )()
        if email:
            fields['emails'].append({'address': email, 'verified': True})

        return data

    @staticmethod
    def deserialize_profile(user, profile, key_prefix='', pop=False):
        """De-serialize user profile fields into concrete model fields."""
        result = {}
        if pop:
            getter = profile.pop
        else:
            getter = profile.get

        def prefixed(name):
            """Return name prefixed by `key_prefix`."""
            return '%s%s' % (key_prefix, name)

        for key in profile.keys():
            val = getter(key)
            if key == prefixed('name'):
                result['full_name'] = val
            else:
                raise ValueError('Bad profile key: %r' % key)
        return result

    @api_endpoint
    def update(self, selector, update, options=None):
        """Update user data."""
        user = get_object(
            self.model, selector['_id'],
            pk=this.request.user.pk,
        )
        profile_update = self.deserialize_profile(
            user, update['$set'], key_prefix='profile.', pop=True,
        )
        if len(update['$set']) != 0:
            raise MeteorError(400, 'Invalid update fields: %r')

        for key, val in profile_update.items():
            setattr(user, key, val)
        user.save()


class LoginPublication(Publication):

    """Meteor Accounts emulation."""

    name = 'meteor.loginServiceConfiguration'

    queries = [
        (Users.model.objects.all(), 'users'),
    ]


class Auth(APIMixin):

    """Meteor Passwords emulation."""

    api_path_prefix = ''  # auth endpoints don't have a common prefix
    user_model = auth.get_user_model()

    def update_subs(self, new_user_id):
        """Update subs to send added/removed for collections with user_rel."""
        for sub in Subscription.objects.filter(connection=this.ws.connection):
            params = loads(sub.params_ejson)
            pub = API.get_pub_by_name(sub.publication)

            # calculate the querysets prior to update
            pre = collections.OrderedDict([
                (col, qs) for col, qs
                in API.sub_unique_objects(sub, params, pub)
            ])

            # save the subscription with the updated user_id
            sub.user_id = new_user_id
            sub.save()

            # calculate the querysets after the update
            post = collections.OrderedDict([
                (col, qs) for col, qs
                in API.sub_unique_objects(sub, params, pub)
            ])

            # first pass, send `added` for objs unique to `post`
            for col_post, qs in post.items():
                try:
                    qs_pre = pre[col_post]
                    qs = qs.exclude(pk__in=qs_pre.order_by().values('pk'))
                except KeyError:
                    # collection not included pre-auth, everything is added.
                    pass
                for obj in qs:
                    this.ws.send(col.obj_change_as_msg(obj, ADDED))

            # second pass, send `removed` for objs unique to `pre`
            for col_pre, qs in pre.items():
                try:
                    qs_post = post[col_pre]
                    qs = qs.exclude(pk__in=qs_post.order_by().values('pk'))
                except KeyError:
                    # collection not included post-auth, everything is removed.
                    pass
                for obj in qs:
                    this.ws.send(col.obj_change_as_msg(obj, REMOVED))

    @staticmethod
    def auth_failed(**credentials):
        """Consistent fail so we don't provide attackers with valuable info."""
        if credentials:
            user_login_failed.send_robust(
                sender=__name__,
                credentials=auth._clean_credentials(credentials),
            )
        raise MeteorError(403, 'Authentication failed.')

    def validated_user_and_session(self, token):
        """Resolve and validate auth token, returns user and session objects."""
        try:
            username, session_key, auth_hash = loads(token.decode('base64'))
        except (ValueError, Error):
            self.auth_failed(token=token)
        try:
            user = self.user_model.objects.get(**{
                self.user_model.USERNAME_FIELD: username,
            })
            user.backend = 'django.contrib.auth.backends.ModelBackend'
        except self.user_model.DoesNotExist:
            self.auth_failed(username=username, token=token)
        if user.get_session_auth_hash() != auth_hash:
            self.auth_failed(username=username, token=token)
        session = SessionStore(
            session_key=session_key,
        )
        if session.get_expiry_date() <= timezone.now():
            self.auth_failed(username=username, token=token)
        return (user, session)

    @staticmethod
    def get_user_token(user, session_key, expiry_date):
        """Return login token info for given user."""
        token = ''.join(
            dumps([
                user.get_username(),
                session_key,
                user.get_session_auth_hash(),
            ]).encode('base64').split('\n')
        )
        return {
            'id': get_meteor_id(user),
            'token': token,
            'tokenExpires': expiry_date,
        }

    @staticmethod
    def check_secure():
        """Check request, return False if using SSL or local connection."""
        if this.request.is_secure():
            return True  # using SSL
        elif this.request.META['REMOTE_ADDR'] in [
                'localhost',
                '127.0.0.1',
        ]:
            return True  # localhost
        raise MeteorError(403, 'Authentication refused without SSL.')

    def get_username(self, user):
        """Retrieve username from user selector."""
        if isinstance(user, basestring):
            return user
        elif isinstance(user, dict) and len(user) == 1:
            [(key, val)] = user.items()
            if key == 'username' or (key == self.user_model.USERNAME_FIELD):
                # username provided directly
                return val
            elif key == 'emails.address':
                email_field = getattr(self.user_model, 'EMAIL_FIELD', 'email')
                if self.user_model.USERNAME_FIELD == email_field:
                    return val  # email is username
                # find username by email
                return self.user_model.objects.values_list(
                    self.user_model.USERNAME_FIELD, flat=True,
                ).get(**{email_field: val})
            elif key in ('id', 'pk'):
                # find username by primary key (ID)
                return self.user_model.objects.values_list(
                    self.user_model.USERNAME_FIELD, flat=True,
                ).get(
                    pk=val,
                )
            else:
                raise MeteorError(400, 'Invalid user lookup: %r' % key)
        else:
            raise MeteorError(400, 'Invalid user expression: %r' % user)

    @staticmethod
    def get_password(password):
        """Return password in plain-text from string/dict."""
        if isinstance(password, basestring):
            # regular Django authentication - plaintext password... but you're
            # using HTTPS (SSL) anyway so it's protected anyway, right?
            return password
        else:
            # Meteor is trying to be smart by doing client side hashing of the
            # password so that passwords are "...not sent in plain text over the
            # wire".  This behaviour doesn't make HTTP any more secure - it just
            # gives a false sense of security as replay attacks and
            # code-injection are both still viable attack vectors for the
            # malicious MITM.  Also as no salt is used with hashing, the
            # passwords are vulnerable to rainbow-table lookups anyway.
            #
            # If you're doing security, do it right from the very outset.  Fors
            # web services that means using SSL and not relying on half-baked
            # security concepts put together by people with no security
            # background.
            #
            # We protest loudly to anyone who cares to listen in the server logs
            # until upstream developers see the light and drop the password
            # hashing mis-feature.
            raise MeteorError(
                400,
                "Outmoded password hashing, run "
                "`meteor add tysonclugg:accounts-secure` to fix.",
            )

    @api_endpoint('createUser')
    def create_user(self, params):
        """Register a new user account."""
        receivers = create_user.send(
            sender=__name__,
            request=this.request,
            params=params,
        )
        if len(receivers) == 0:
            raise MeteorError(501, 'Handler for `create_user` not registered.')
        user = receivers[0][1]
        user = auth.authenticate(
            username=user.get_username(), password=params['password'],
        )
        auth.login(this.request, user)
        self.update_subs(user.pk)
        return self.get_user_token(
            user=user,
            session_key=this.request.session.session_key,
            expiry_date=this.request.session.get_expiry_date(),
        )

    @api_endpoint
    def logout(self):
        """Logout current user."""
        auth.logout(this.request)
        self.update_subs(None)

    @api_endpoint
    def login(self, params):
        """Login either with resume token or password."""
        if 'password' in params:
            return self.login_with_password(params)
        elif 'resume' in params:
            return self.login_with_resume_token(params)
        else:
            self.auth_failed(**params)

    def login_with_password(self, params):
        """Authenticate using credentials supplied in params."""
        # never allow insecure login
        self.check_secure()

        username = self.get_username(params['user'])
        password = self.get_password(params['password'])

        user = auth.authenticate(username=username, password=password)
        if user is not None:
            # the password verified for the user
            if user.is_active:
                auth.login(this.request, user)
                self.update_subs(user.pk)
                this.request.session.save()
                return self.get_user_token(
                    user=user,
                    session_key=this.request.session.session_key,
                    expiry_date=this.request.session.get_expiry_date(),
                )

        # Call to `authenticate` was unable to verify the username and password.
        # It will have sent the `user_login_failed` signal, no need to pass the
        # `username` argument to auth_failed().
        self.auth_failed()

    def login_with_resume_token(self, params):
        """
        Login with existing resume token.

        Either the token is valid and the user is logged in, or the token is
        invalid and a non-specific ValueError("Login failed.") exception is
        raised - don't be tempted to give clues to attackers as to why their
        logins are invalid!
        """
        # never allow insecure login
        self.check_secure()

        # pull the username, session_key and session_auth_hash from the token
        user, session = self.validated_user_and_session(params['resume'])

        auth.login(this.request, user)
        self.update_subs(user.pk)
        this.request.session.save()
        return self.get_user_token(
            user=user,
            session_key=session.session_key,
            expiry_date=session.get_expiry_date(),
        )

    @api_endpoint('changePassword')
    def change_password(self, params):
        """Change password."""
        user = auth.authenticate(
            username=this.request.user.get_username(),
            password=self.get_password(params['oldPassword']),
        )
        if user is None:
            self.auth_failed()
        else:
            user.set_password(self.get_password(params['newPassword']))
            user.save()
            password_changed.send(
                sender=__name__,
                request=this.request,
                user=user,
            )

    @api_endpoint('forgotPassword')
    def forgot_password(self, params):
        """Request password reset email."""
        username = self.get_username(params['user'])
        try:
            user = self.user_model.objects.get(**{
                self.user_model.USERNAME_FIELD: username,
            })
        except self.user_model.DoesNotExist:
            self.auth_failed()

        expiry_date = this.request.session.get_expiry_date()
        token = self.get_user_token(
            user=user, session_key=this.request.session.session_key,
            expiry_date=expiry_date,
        )

        forgot_password.send(
            sender=__name__,
            user=user,
            token=token,
            request=this.request,
            expiry_date=expiry_date,
        )

    @api_endpoint('resetPassword')
    def reset_password(self, params):
        """Reset password using a token received in email then logs user in."""
        user, _ = self.validated_user_and_session(params['token'])
        user.set_password(params['newPassword'])
        user.save()
        auth.login(this.request, user)
        self.update_subs(user.pk)


API.register([Users, LoginPublication, Auth])



from django.db.migrations.operations.base import Operation


class TruncateOperation(Operation):

    """Truncate (delete all rows) from the models specified."""

    def __init__(self, forwards=None, backwards=None):
        """Accept model names which are to be migrated."""
        self.truncate_forwards = forwards or []
        self.truncate_backwards = backwards or []

    def truncate(self, app_label, schema_editor, models):
        """Truncate tables."""
        for model_name in models:
            model = '%s_%s' % (app_label, model_name)
            schema_editor.execute(
                'TRUNCATE TABLE %s RESTART IDENTITY CASCADE' % (
                    model.lower(),
                ),
            )

    def state_forwards(self, app_label, state):
        """Mutate state to match schema changes."""
        pass  # Truncate doesn't change schema.

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        """Use schema_editor to apply any forward changes."""
        self.truncate(app_label, schema_editor, self.truncate_forwards)

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        """Use schema_editor to apply any reverse changes."""
        self.truncate(app_label, schema_editor, self.truncate_backwards)

    def describe(self):
        """Describe what the operation does in console output."""
        return "Truncate tables"



"""Django DDP Server views."""
from __future__ import print_function, absolute_import
from ejson import dumps
from django.apps import apps
from django.http import HttpResponse
from django.views.generic import View


STAR_JSON_SETTING_NAME = 'METEOR_STAR_JSON'


class MeteorView(View):

    """Django DDP Meteor server view."""

    http_method_names = ['get', 'head']

    app = None
    runtime_config = None

    def __init__(self, **kwargs):
        """Initialisation for Django DDP server view."""
        super(MeteorView, self).__init__(**kwargs)
        self.app = apps.get_app_config('server')

    def get(self, request, path):
        """Return HTML (or other related content) for Meteor."""
        if path == '/meteor_runtime_config.js':
            config = {
                'DDP_DEFAULT_CONNECTION_URL': request.build_absolute_uri('/'),
                'ROOT_URL': request.build_absolute_uri('/'),
                'ROOT_URL_PATH_PREFIX': '',
            }
            config.update(self.runtime_config)
            return HttpResponse(
                '__meteor_runtime_config__ = %s;' % dumps(config),
                content_type='text/javascript',
            )
        try:
            file_path, content_type = self.app.url_map[path]
            with open(file_path, 'r') as content:
                return HttpResponse(
                    content.read(),
                    content_type=content_type,
                )
        except KeyError:
            print(path)
            return HttpResponse(self.app.html)
            # raise Http404



#!/usr/bin/env python
"""Entry point for Django DDP test project."""
import os
import sys


if __name__ == "__main__":
    os.environ['DJANGO_SETTINGS_MODULE'] = 'dddp.test.test_project.settings'

    from dddp import greenify
    greenify()

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dddp', '0002_auto_20150408_0321'),
    ]

    operations = [
        migrations.AlterField(
            model_name='objectmapping',
            name='object_id',
            field=models.CharField(max_length=255),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dddp', '0006_auto_20150428_2245'),
    ]

    operations = [
        migrations.RenameField(
            model_name='subscriptioncollection',
            old_name='collection_class',
            new_name='collection_name',
        ),
        migrations.RenameField(
            model_name='subscriptioncollection',
            old_name='name',
            new_name='model_name',
        ),
    ]



"""Django DDP PostgreSQL Greenlet."""

from __future__ import absolute_import

import ejson
import gevent
import gevent.queue
import gevent.select
import psycopg2  # green
import psycopg2.extensions
from geventwebsocket.logging import create_logger


class PostgresGreenlet(gevent.Greenlet):

    """Greenlet for multiplexing database operations."""

    def __init__(self, conn, debug=False):
        """Prepare async connection."""
        super(PostgresGreenlet, self).__init__()
        self.logger = create_logger(__name__, debug=debug)

        # queues for processing incoming sub/unsub requests and processing
        self.connections = {}
        self._stop_event = gevent.event.Event()

        # connect to DB in async mode
        conn.allow_thread_sharing = True
        self.connection = conn
        self.conn_params = conn.get_connection_params()
        self.conn_params.update(
            async=True,
        )
        self.conn = psycopg2.connect(**self.conn_params)
        self.poll()  # wait for conneciton to start
        self.cur = self.conn.cursor()

    def _run(self):  # pylint: disable=method-hidden
        """Spawn sub tasks, wait for stop signal."""
        gevent.spawn(self.process_conn)
        self._stop_event.wait()

    def stop(self):
        """Stop subtasks and let run() finish."""
        self._stop_event.set()

    def process_conn(self):
        """Subtask to process NOTIFY async events from DB connection."""
        self.cur.execute('LISTEN "ddp";')
        while not self._stop_event.is_set():
            gevent.select.select([self.conn], [], [], timeout=None)
            self.poll()

    def poll(self):
        """Poll DB socket and process async tasks."""
        while 1:
            state = self.conn.poll()
            if state == psycopg2.extensions.POLL_OK:
                while self.conn.notifies:
                    notify = self.conn.notifies.pop()
                    self.logger.info(
                        "Got NOTIFY (pid=%d, payload=%r)",
                        notify.pid, notify.payload,
                    )
                    data = ejson.loads(notify.payload)
                    sender = data.pop('_sender', None)
                    tx_id = data.pop('_tx_id', None)
                    for connection_id in data.pop('_connection_ids'):
                        try:
                            ws = self.connections[connection_id]
                        except KeyError:
                            continue  # connection not in this process
                        if connection_id == sender:
                            ws.send(data, tx_id=tx_id)
                        else:
                            ws.send(data)
                break
            elif state == psycopg2.extensions.POLL_WRITE:
                gevent.select.select([], [self.conn.fileno()], [])
            elif state == psycopg2.extensions.POLL_READ:
                gevent.select.select([self.conn.fileno()], [], [])
            else:
                self.logger.warn('POLL_ERR: %s', state)



from dddp.api import API, Collection, Publication
from dddp.test.django_todos import models


class Task(Collection):
    model = models.Task


class Tasks(Publication):
    queries = [
        models.Task.objects.all(),
    ]

API.register([
    Task,
    Tasks,
])



"""Django DDP app config."""

from __future__ import print_function

from django.apps import AppConfig
from django.conf import settings, ImproperlyConfigured

from dddp import autodiscover


class DjangoDDPConfig(AppConfig):

    """Django app config for django-ddp."""

    api = None
    name = 'dddp'
    verbose_name = 'Django DDP'
    _in_migration = False

    def ready(self):
        """Initialisation for django-ddp (setup lookups and signal handlers)."""
        if not settings.DATABASES:
            raise ImproperlyConfigured('No databases configured.')
        for (alias, conf) in settings.DATABASES.items():
            if conf['ENGINE'] != 'django.db.backends.postgresql_psycopg2':
                raise ImproperlyConfigured(
                    '%r uses %r: django-ddp only works with PostgreSQL.' % (
                        alias, conf['backend'],
                    )
                )
        self.api = autodiscover()
        self.api.ready()



from django.conf import settings
from django.conf.urls import patterns, include, url
from django.contrib import admin

urlpatterns = patterns('',
    # Examples:
    # url(r'^$', 'dddp.test.test_project.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),

    url(r'^admin/', include(admin.site.urls)),
    url(
        r'^static/(?P<path>.*)$',
        'django.views.static.serve',
        {
            'document_root': settings.STATIC_ROOT,
            'show_indexes': False,
        },
    ),
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dddp', '0003_auto_20150413_1328'),
    ]

    operations = [
        migrations.AddField(
            model_name='connection',
            name='server_addr',
            field=models.CharField(default='', max_length=255),
            preserve_default=False,
        ),
    ]
